/*  "segs.scm" WB-tree File Based Associative String Data Base System. */
/*  Copyright (C) 1991, 1992, 1993, 2000, 2007 Free Software Foundation, Inc. */
/*  */
/*  This program is free software: you can redistribute it and/or modify */
/*  it under the terms of the GNU Lesser General Public License as */
/*  published by the Free Software Foundation, either version 3 of the */
/*  License, or (at your option) any later version. */
/*  */
/*  This program is distributed in the hope that it will be useful, but */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU */
/*  Lesser General Public License for more details. */
/*  */
/*  You should have received a copy of the GNU Lesser General Public */
/*  License along with this program.  If not, see */
/*  <http://www.gnu.org/licenses/>. */

/*  routines in this file return success codes */


#include "wbsys.h"

/* The @var{bsiz} of a segment (given in call to @code{make_seg}) */
/* is a parameter crucial for performance; balancing CPU time traversing */
/* blocks with file-system latency.  @var{bsiz} should be an */
/* integer multiple of the file-system block size. */
/* In the 1990s our nominal @var{bsiz} was 2.kiB; now it should */
/* probably be 4.kiB, 8.kiB, or 16.kiB. */

int seg_cntr = 4;

SEGD *seg_chain = 0;

LCK seg_chain_lock = {0, -3, PTHREAD_MUTEX_INITIALIZER};
LCK *seg_chain_lck = &seg_chain_lock;
/*  Create segment-descriptor and splice into seg-chain. */

SEGD *new_seg(name)
     unsigned char *name;
{
  if (!(lck_tab))
    {
      dprintf((diagout,
	 ">>>>ERROR<<<< %s: WB not initialized (need to call %s)\n",
	 name,
	 "init_wb"));
      return 0;
    }
  else {
    lck(seg_chain_lck);
    {
      SEGD *seg = new_segd(seg_cntr);
      if (seg_chain)
	{
	  SEGD *prev_seg = seg_prv(seg_chain);
	  seg_set_nxt(prev_seg, seg);
	  seg_set_prv(seg, prev_seg);
	  seg_set_nxt(seg, seg_chain);
	  seg_set_prv(seg_chain, seg);
	}
      else {
	seg_set_nxt(seg, seg);
	seg_set_prv(seg, seg);
      }
      seg_set_flc(seg, calloc(flc_len, (sizeof (void *))));
      seg_set_flc_len(seg, 0);
      seg_cntr = 1+(seg_cntr);
      seg_chain = seg;
      unlck(seg_chain_lck);
      return seg;
    }
  }
}

/*  lock must already be held on seg-chain-lck */

void free_seg(seg)
     SEGD *seg;
{
  if (!(seg_chain))
    {
      dprintf((diagout, ">>>>ERROR<<<< %s: %s already empty!\n", "free_seg", "seg_chain"));
      return;
    }
  else if ((seg_prv(seg))==(seg))
    {
      if ((seg)!=(seg_chain))
	dprintf((diagout, ">>>>ERROR<<<< %s: corrupted %s.\n", "free_seg", "seg_chain"));
      seg_chain = 0;
      return;
    }
  else {
    SEGD *next_seg = seg_nxt(seg);
    SEGD *prev_seg = seg_prv(seg);
    seg_set_nxt(prev_seg, next_seg);
    seg_set_prv(next_seg, prev_seg);
    if ((seg)==(seg_chain))
      seg_chain = (seg_chain)==(next_seg)
	?0
	:next_seg;
      return;
  }
}


int check_file_P(filename)
     unsigned char *filename;
{
  int file = blkio_open_read_only_file(filename, 0x80);
  unsigned char tblk[0x19];
  subbytes_read(tblk, 0, 0x19, file);
  {
    int allok_P = (!(str2long(tblk, 0)))
    && (!(str2long(tblk, 4)))
    && (!(str2long(tblk, 8)))
    && (0x77==((((unsigned char*)(tblk))[0x17])))
    && (0x62==((((unsigned char*)(tblk))[0x18])));
    blkio_file_close(file);
    return allok_P;
  }
}

/*  */
/*  */
/* Opens the database file @var{filename} and returns a @dfn{seg}, false otherwise. */
/* The database will be read-only if the @var{mutable_P} argument is false.  It will */
/* be read-write if the @var{mutable_P} argument is true. */

SEGD *open_seg(filename, mutable_P)
     unsigned char *filename;
     int mutable_P;
{
  int bsiz = 3*0x80;
  if ((bsiz)>(blk_size))
    {
      dprintf((diagout, ">>>>ERROR<<<< unsupported %s %d > %d\n", "bsiz", bsiz, blk_size));
      return 0;
    }
  else if (!(check_file_P(filename)))
    return 0;
  else {
    SEGD *seg = new_seg("open_seg");
    char *reason;
    if (!(seg))
      return seg;
    else {
      int file = mutable_P
	?blkio_open_modify_file(filename, bsiz)
	:blkio_open_read_only_file(filename, bsiz);
L_loop:
      if (mutable_P
	    ?output_port_P(file)
	    :input_port_P(file))
	{
	  seg_set_port(seg, file);
	  seg_set_bsiz(seg, bsiz);
	  seg_set_used(seg, 2);
	  seg_set_str(seg, filename);
	  seg_set_flc_len(seg, mutable_P
	      ? -1
	      : -2);
	  {
	    HAND *han = seg_rt_han(seg);
	    unsigned char tmp_str[5];
	    if (err_P(bt_open(seg, 0, han, (wcb_sap)+(wcb_sar))))
	      {
		reason = "bt-open 0";
		goto L_errout;
	      }
	    else if (2!=(bt_get(han, bsiz_byts, 4, tmp_str)))
	      {
		reason = "BSIZ";
		goto L_errout;
	      }
	    else if ((bsiz)!=(str2short(tmp_str, 0)))
	      {
		blkio_file_close(file);
		bsiz = str2short(tmp_str, 0);
		if ((bsiz)>(blk_size))
		  {
		    reason = "BSIZ too big.";
		    goto L_errout;
		  }
		else {
		  file = mutable_P
		    ?blkio_open_modify_file(filename, bsiz)
		    :blkio_open_read_only_file(filename, bsiz);
		  goto L_loop;
		}
	      }
	    else if (4!=(bt_get(han, used_byts, 4, tmp_str)))
	      {
		reason = "USED";
		goto L_errout;
	      }
	    else {
	      seg_set_used(seg, str2long(tmp_str, 0));
	      if (5!=(bt_get(han, fld_byts, 3, tmp_str)))
		{
		  reason = "FLD";
		  goto L_errout;
		}
	      else if (err_P(bt_open(seg, str2long(tmp_str, 1), seg_fl_han(seg), wcb_sar)))
		{
		  reason = "FLC";
		  goto L_errout;
		}
	      else {
		if ((han_typ(seg_fl_han(seg)))!=(frl_typ))
		  dprintf((diagout, "Older type freelist - still supported.\n"));
		han_set_wcb(seg_fl_han(seg), wcb_sar);
		return seg;
	      }
	    }
	  }
	}
      else {
	if (mutable_P
	      ?input_port_P(file)
	      :output_port_P(file))
	  blkio_file_close(file);
	dprintf((diagout, ">>>>ERROR<<<< could not open file %s\n", filename));
	free_seg(seg);
	return 0;
      }
    }
L_errout:
    dprintf((diagout, ">>>>ERROR<<<< not a database %s %s\n", filename, reason));
    if (seg)
      {
	blkio_file_close(seg_port(seg));
	do_seg_buffers(seg, recycle_ent);
	seg_set_port(seg, 0);
	seg_set_str(seg, 0);
	seg_set_used(seg, 0);
	free_seg(seg);
      }
    return 0;
  }
}

/*  */
/* Closes database segment @var{seg} and the file containing it.  If @var{hammer_P} is */
/* NULL, then if there are any problems freeing buffers, then the close */
/* is aborted.  A status code is returned. */

int close_seg(seg, hammer_P)
     SEGD *seg;
     int hammer_P;
{
  return close_segd(seg, hammer_P, !(!0));
}


int close_segd(seg, hammer_P, all_P)
     SEGD *seg;
     int hammer_P;
     int all_P;
{
  if (!(all_P))
    lck(seg_chain_lck);
  if ((!(seg_str(seg)))
      || (!(seg_used(seg))))
    {
      if (!(all_P))
	unlck(seg_chain_lck);
      return 0;
    }
  else {
    flush_flc(seg, 5);
    /*leave only enough blocks to fit in flc in superblock.*/
    if ((seg_flc_len(seg))>=0)
      {
	unsigned char tmp_str[0x14];
	{
	  int i =  -1+(seg_flc_len(seg));
	  while (!(0 > (i))) {
	    long2str(tmp_str, 4*(i), seg_flc(seg)[i]);
	    {
	      i =  -1+(i);
	    }
	  }
	}
	bt_put(seg_rt_han(seg), flc_byts, 3, tmp_str, 4*(seg_flc_len(seg)));
	seg_set_flc_len(seg,  -1);
      }
    {
      int ans = do_seg_buffers(seg, flush_buffer);
      if ((success_P(ans))
	  || (hammer_P))
	{
	  if (!(success_P(ans)))
	    ans = notpres;
	  do_seg_buffers(seg, purge_buffer);
	  bt_close(seg_rt_han(seg));
	  bt_close(seg_fl_han(seg));
	  blkio_file_close(seg_port(seg));
	  seg_set_port(seg, 0);
	  seg_set_str(seg, 0);
	  free(seg_flc(seg));
	  seg_set_flc(seg, 0);
	  seg_set_flc_len(seg, 0);
	  seg_set_used(seg, 0);
	  free_seg(seg);
	}
      if (!(all_P))
	unlck(seg_chain_lck);
      return ans;
    }
  }
}


unsigned char db_version_str[] = "wb-2a3";

unsigned char db_authors_str[] = "A. Jaffer, J. Finger, R. Zito-Wolf";

/*  */
/*  */
/* The integer @var{bsiz} specifies the size of B-tree blocks.  @var{bsiz} should be an */
/* integer multiple of the file-system block size.  Nominal value is */
/* 4096. */
/*  */
/* @code{make_seg} returns an open new empty mutable database named backed by file */
/* @var{filename} if successful; otherwise false is returned. */

SEGD *make_seg(filename, bsiz)
     unsigned char *filename;
     int bsiz;
{
  SEGD *seg = new_seg("make_seg");
  if (!(seg))
    return seg;
  else if ((bsiz)>(blk_size))
    {
      dprintf((diagout, ">>>>ERROR<<<< unsupported %s %d > %d\n", "bsiz", bsiz, blk_size));
      unlck(seg_lck(seg));
      return 0;
    }
  else {
    int file = blkio_create_file(filename, bsiz);
    if (output_port_P(file))
      {
	seg_set_port(seg, file);
	seg_set_bsiz(seg, bsiz);
	seg_set_used(seg, 3);
	seg_set_str(seg, filename);
	seg_set_flc_len(seg,  -1);
	{
	  HAND *rt_han = seg_rt_han(seg);
	  HAND *han = seg_fl_han(seg);
	  unsigned char tmp_str[5];
	  if ((bt_open_new(seg, 0, rt_han, (wcb_sap)+(wcb_sar), dir_typ))
	      && (bt_open_new(seg, 1, han, wcb_sap, dir_typ))
	      && (bt_open_new(seg, 2, han, wcb_sar, frl_typ)))
	    {
	      bt_put(rt_han, no_byts, 0, db_version_str, strlen(db_version_str));
	      long2str(tmp_str, 0, seg_used(seg));
	      bt_put(rt_han, used_byts, 4, tmp_str, 4);
	      short2str(tmp_str, 0, seg_bsiz(seg));
	      bt_put(rt_han, bsiz_byts, 4, tmp_str, 2);
	      tmp_str[0] = 4;
	      long2str(tmp_str, 1, 1);
	      bt_put(rt_han, root_byts, 4, tmp_str, 5);
	      long2str(tmp_str, 1, 2);
	      bt_put(rt_han, fld_byts, 3, tmp_str, 5);
	      bt_put(rt_han, flc_byts, 3, no_byts, 0);
	      if ((bsiz)>0x80)
		bt_put(rt_han, authors_byts, 7, db_authors_str, strlen(db_authors_str));
	      return seg;
	    }
	  else {
	    dprintf((diagout,
	 ">>>>ERROR<<<< couldn't allocate ents for file %s\n",
	 filename));
	    unlck(seg_lck(seg));
	    close_seg(seg, !0);
	    return 0;
	  }
	}
      }
    else {
      dprintf((diagout,
	 ">>>>ERROR<<<< couldn't create new file %s\n",
	 filename));
      unlck(seg_lck(seg));
      return 0;
    }
  }
}

/* @noindent The write-control-bits argument (@var{wcb}) to these */
/* functions controls the latency of updates to the file after various */
/* operations.  These bits are defined as follows: */
/*  */
/* @multitable @columnfractions .1 .3 .6 */
/* @item @var{value} */
/* @tab C-name */
/* @tab Meaning */
/* @item 1 */
/* @tab wcb_sap */
/* @tab save block after PUTs */
/* @cindex wcb_sap */
/* @item 2 */
/* @tab wcb_sar */
/* @tab save block after REMOVEs */
/* @cindex wcb_sar */
/* @item 4 */
/* @tab wcb_sac */
/* @cindex wcb_sac */
/* @tab force block save after cached block changes */
/* @item 8 */
/* @tab wcb_fac */
/* @tab flush buffer entirely after cached block changes (not currently */
/* implemented) */
/* @cindex wcb_fac */
/* @end multitable */
/* Opens bt-handle @var{han} to seg number @var{seg}, block number @var{blk_num}, and */
/* returns the type of the block.  If no such block exists or is not a */
/* root block, then a (negative) status code is returned. */

int bt_open(seg, blk_num, han, wcb)
     SEGD *seg;
     long blk_num;
     HAND *han;
     int wcb;
{
  if ((seg)
      && (seg_str(seg)))
    {
      ENTRY *ent = get_ent(seg, blk_num, accread);
      if (!(ent))
	return argerr;
      else if (!(root_P(ent_blk(ent))))
	{
	  release_ent(ent, accread);
	  dprintf((diagout, ">>>>ERROR<<<< %s: not a root %ld:%ld\n", "bt_open", seg_id(seg), blk_num));
	  return argerr;
	}
      else {
	bt_open_init_han(han, ent, wcb);
	release_ent(ent, accread);
	return han_typ(han);
      }
    }
  else return argerr;
}


void bt_open_init_han(han, ent, wcb)
     HAND *han;
     ENTRY *ent;
     int wcb;
{
  int typ = blk_typ(ent_blk(ent));
  SEGD *seg = ent_seg(ent);
  long blk_num = ent_id(ent);
  han_set_seg(han, seg);
  han_set_num(han, blk_num);
  han_set_last(han, blk_num);
  han_set_typ(han, typ);
  if ((typ)==(dir_typ))
    wcb = (wcb)|((wcb_sap)+(wcb_sar));
  han_set_wcb(han, wcb);
  return;
}

/*  Does calling GET-BUK-WAIT admit the possiblity of deadly embrace? */

int bt_open_new(seg, blk_num, han, wcb, typ)
     SEGD *seg;
     long blk_num;
     HAND *han;
     int wcb;
     int typ;
{
  if (0 > (blk_num))
    {
      dprintf((diagout, ">>>>ERROR<<<< negative block number %ld\n", blk_num));
      return 0;
    }
  else {
    ENTRY *ent = try_get_free_ent(0,  -1);
    while (!(ent)) {
      {
	ent = try_get_free_ent(0,  -1);
      }
    }
    init_leaf_blk(ent_blk(ent), blk_num, typ);
    blk_set_time(ent_blk(ent), time(0L));
    ent_set_next(ent, get_buk_wait(seg_id(seg), blk_num));
    set_buk(seg_id(seg), blk_num, ent);
    ent_set_acc(ent, accwrite);
    ent_set_seg(ent, seg);
    ent_set_id(ent, blk_num);
    ent_set_age(ent,  -127);
    ent_set_dty(ent, !0);
    ent_set_pus(ent, 0);
    ent_set_ref(ent, 1);
    rel_buk(seg_id(seg), blk_num);
    bt_open_init_han(han, ent, wcb);
    ents_ent_write(ent);
    release_ent(ent, accwrite);
    return !0;
  }
}

/* Creates a new root block in seg @var{seg} of type @var{typ}, opens bt-handle */
/* @var{han} to it, and returns a status code.  If @var{seg} has insufficient room to */
/* create a new tree, then the @var{noroom} status code is returned. */
/*  */
/* @code{bt_create} can be used to create temporary b-trees.  Temporary trees will be */
/* be reclaimed by check program after system crashes.  In order to */
/* make a tree persistent, add it to a directory (tree). */

int bt_create(seg, typ, han, wcb)
     SEGD *seg;
     int typ;
     HAND *han;
     int wcb;
{
  ENTRY *ent = create_new_blk_ent(seg);
  if (ent)
    {
      long blk_num = ent_id(ent);
      init_leaf_blk(ent_blk(ent), blk_num, typ);
      ent_set_pus(ent, 0);
      bt_open_init_han(han, ent, wcb);
      ents_ent_write(ent);
      release_ent(ent, accwrite);
      return success;
    }
  else return noroom;
}

/* Closes bt-handle @var{han} and returns @var{SUCCESS}. */
/*  */
/* Currently, @code{bt_close} has no effect other than to clear @var{han}. */

int bt_close(han)
     HAND *han;
{
  han_set_seg(han, 0);
  han_set_num(han, 0);
  han_set_typ(han, 0);
  han_set_last(han, 0);
  return success;
}
