/*  "handle.scm" WB-tree File Based Associative String Data Base System. */
/*  Copyright (C) 1991, 1992, 1993, 2000, 2007 Free Software Foundation, Inc. */
/*  */
/*  This program is free software: you can redistribute it and/or modify */
/*  it under the terms of the GNU Lesser General Public License as */
/*  published by the Free Software Foundation, either version 3 of the */
/*  License, or (at your option) any later version. */
/*  */
/*  This program is distributed in the hope that it will be useful, but */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU */
/*  Lesser General Public License for more details. */
/*  */
/*  You should have received a copy of the GNU Lesser General Public */
/*  License along with this program.  If not, see */
/*  <http://www.gnu.org/licenses/>. */

/*  routines in this file return success codes */



package wb;

import static wb.Ents.*;
import static wb.Prev.*;
import static wb.Stats.*;
import static wb.Blink.*;
import static wb.Del.*;
import static wb.Scan.*;
import static wb.Wbdefs.*;
import static wb.Wbsys.*;
import static wb.BooleanMethods.*;
import static wb.Pkt.*;
import static wb.Ent.*;
import static wb.Han.*;
public class Handle {


public static boolean cleverCacheEnable_P = (true);

/* Note: most of the data-manipulating commands here */
/* can return @var{notpres}, with the followng meanings: */
/* @cindex notpres */
/*  */
/* @multitable @columnfractions .2 .8 */
/* @item @code{bt-get} */
/* @tab no such key. */
/* @item @code{bt-next} */
/* @tab no NEXT key (ie, key given was LAST key). */
/* @item @code{bt-prev} */
/* @tab no PREV key (ie, key given was FIRST key). */
/* @item @code{bt-rem} */
/* @tab KEY was not found. */
/* @c @item @code{bt-rem-range} */
/* @c @tab ?? */
/* @item @code{bt-put} */
/* @tab NOT USED (could be symmetric w/WRITE). */
/* @item @code{bt-write} */
/* @tab key WAS found, so no write done. */
/* @end multitable */
/* @var{keyStr} is a string of length @var{kLen}.  @code{btGet} stores into the string @var{ansStr} the */
/* value associated with @var{keyStr} in tree @var{han}.  @code{btGet} returns the length of the */
/* string stored into @var{ansStr} or an error code. */

public static int btGet(Han han,byte []keyStr,int kLen,byte []ansStr)
{
  int []pkt = new int[pktSize];
  Ent ent = null;
  ent = chainFindEnt(han, accread, keyStr, kLen, pkt);
  if (!(a2b(ent)))
    {
      getFct = 1+(getFct);
      return unkerr;
    }
  else if ((pkt_MatchType(pkt))!=(match))
    {
      getCt = 1+(getCt);
      releaseEnt(ent, accread);
      return notpres;
    }
  else {
    int alen = getThisVal(ent_Blk(ent), pkt_MatchPos(pkt), ansStr);
    getCt = 1+(getCt);
    releaseEnt(ent, accread);
    return alen;
  }
}

/* @var{keyStr} is a string of length @var{kLen}.  @code{btNext} stores into the string @var{ansStr} the */
/* next key after @var{keyStr} in tree @var{han}.  @code{btNext} returns the length of the string */
/* stored into @var{ansStr} or an error code. */

public static int btNext(Han han,byte []keyStr,int kLen,byte []ansStr)
{
  int []pkt = new int[pktSize];
  Ent ent = null;
  ent = chainFindEnt(han, accread, keyStr, kLen, pkt);
  if (!(a2b(ent)))
    {
      nextFct = 1+(nextFct);
      return unkerr;
    }
  else {
    nextCt = 1+(nextCt);
    {
      int res = chainNext(ent, keyStr, kLen, ansStr, pkt);
      if (cleverCacheEnable_P)
	han_SetLast(han, pkt_BlkToCache(pkt));
      return res;
    }
  }
}

/* @var{keyStr} is a string of length @var{kLen}.  @code{btPrev} stores into the string @var{ansStr} the */
/* last key before @var{keyStr} in tree @var{han}.  @code{btPrev} returns the length of the string */
/* stored into @var{ansStr} or an error code. */

public static int btPrev(Han han,byte []keyStr,int kLen,byte []ansStr)
{
  int []pkt = new int[pktSize];
  Ent ent = null;
  ent = chainFindPrevEnt(han, accread, keyStr, kLen, pkt);
  if (a2b(ent))
    ent = prev_PrevKEnt(ent, keyStr, kLen, leaf, pkt);
  if (!(a2b(ent)))
    {
      prevFct = 1+(prevFct);
      return unkerr;
    }
  else {
    prevCt = 1+(prevCt);
    if (0==(pkt_MatchPos(pkt)))
      {
	releaseEnt(ent, accread);
	return notpres;
      }
    else {
      int kLen2 = reconThisKey(ent_Blk(ent), pkt_MatchPos(pkt), ansStr, 0, 0x100);
      han_SetLast(han, ent_Id(ent));
      releaseEnt(ent, accread);
      return kLen2;
    }
  }
}


/* @var{keyStr} is a string of length @var{kLen}.  @code{btRem} stores into the string @var{ansStr} the */
/* value associated with @var{keyStr} in tree @var{han}; then removes that association */
/* from tree @var{han}.  @code{btRem} returns the length of the string stored into @var{ansStr} or */
/* an error code. */
/*  */
/* If @var{ansStr} is 0, @code{btRem} removes the @var{keyStr} association from tree @var{han} and returns */
/* @var{SUCCESS} if successful; an error code if not. */

public static int btRem(Han han,byte []keyStr,int kLen,byte []ansStr)
{
  int []pkt = new int[pktSize];
  Ent ent = null;
  if ((kLen)<0)
    {
      dprintf(">>>>ERROR<<<< "+("btRem")+": bad length string "+(kLen)+"\n");
      return argerr;
    }
  else {
    ent = chainFindEnt(han, accwrite, keyStr, kLen, pkt);
    if (a2b(ent))
      {
	remCt = 1+(remCt);
	{
	  int ans = del_ChainRem(ent, keyStr, kLen, ansStr, pkt, han_Wcb(han));
	  releaseEnt(ent, accwrite);
	  return ans;
	}
      }
    else {
      remFct = 1+(remFct);
      return unkerr;
    }
  }
}

/*  To make possible bounded-time operation, @0 will purge at most */
/*  BLK-LIMIT blocks at a time; passing BLK-LIMIT of -1 imposes no */
/*  limit. */

/* @var{keyStr} must be a maximum-length (256 byte) string containing a key */
/* @var{kLen} bytes long.  @var{key2Str} is a string of length @var{k2Len}. */
/*  */
/* @code{btRemRange} removes [@var{keyStr} @dots{} @var{key2Str}) and their values.  If @var{key2Str} <= @var{keyStr} no */
/* deletion will occur (even if @var{keyStr} is found).  @code{btRemRange} returns SUCCESS if */
/* the operation is complete, an error status code if not. */

public static int btRemRange(Han han,byte []keyStr,int kLen,byte []key2Str,int k2Len)
{
  int []respkt = new int[pktSize];
  pkt_SetSkeyCount(respkt, 0);
  return btScan(han, remScan, keyStr, kLen, key2Str, k2Len, null, null, respkt,  -1);
}

/* @var{keyStr} is a string of length @var{kLen}.  @var{valStr} is a string of length @var{vLen}.  @code{btPut} */
/* makes the value associated with @var{keyStr} be @var{valStr} in tree @var{han}.  @code{btPut} returns a */
/* status code for the operation. */

public static int btPut(Han han,byte []keyStr,int kLen,byte []valStr,int vLen)
{
  Ent ent = null;
  int []pkt = new int[pktSize];
  if (((vLen)>0xff
      || (kLen)>0xff
      || (kLen)<0))
    return argerr;
  else {
    ent = chainFindEnt(han, accwrite, keyStr, kLen, pkt);
    if (a2b(ent))
      {
	boolean res_P = chainPut(ent, keyStr, kLen, valStr, vLen, pkt, null, han_Wcb(han));
	if (res_P)
	  {
	    if (cleverCacheEnable_P)
	      han_SetLast(han, pkt_BlkToCache(pkt));
	    putCt = 1+(putCt);
	    return success;
	  }
	else {
	  putFct = 1+(putFct);
	  return unkerr;
	}
      }
    else return unkerr;
  }
}

/* @var{keyStr} is a string of length @var{kLen}.  @var{valStr} is a string of length @var{vLen}.  If */
/* @var{han} currently contains an association for @var{keyStr}, then @code{btWrite} does not modify */
/* the tree and returns the @var{notpres} status code. */
/*  */
/* Otherwise, @code{btWrite} makes the value associated with @var{keyStr} be @var{valStr} in tree @var{han}. */
/* @code{btWrite} returns a status code for the operation. */

public static int btWrite(Han han,byte []keyStr,int kLen,byte []valStr,int vLen)
{
  Ent ent = null;
  int []pkt = new int[pktSize];
  if (((vLen)>0xff
      || (kLen)>0xff
      || (kLen)<0))
    return argerr;
  else {
    ent = chainFindEnt(han, accwrite, keyStr, kLen, pkt);
    if (a2b(ent))
      if ((pkt_MatchType(pkt))==(match))
	{
	  releaseEnt(ent, accwrite);
	  return notpres;
	}
      else {
	boolean res_P = chainPut(ent, keyStr, kLen, valStr, vLen, pkt, null, han_Wcb(han));
	if (res_P)
	  {
	    if (cleverCacheEnable_P)
	      han_SetLast(han, pkt_BlkToCache(pkt));
	    putCt = 1+(putCt);
	    return success;
	  }
	else {
	  putFct = 1+(putFct);
	  return unkerr;
	}
      }
    else return unkerr;
  }
}
	

}
