/*  "segs.scm" WB-tree File Based Associative String Data Base System. */
/*  Copyright (C) 1991, 1992, 1993, 2000, 2007 Free Software Foundation, Inc. */
/*  */
/*  This program is free software: you can redistribute it and/or modify */
/*  it under the terms of the GNU Lesser General Public License as */
/*  published by the Free Software Foundation, either version 3 of the */
/*  License, or (at your option) any later version. */
/*  */
/*  This program is distributed in the hope that it will be useful, but */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU */
/*  Lesser General Public License for more details. */
/*  */
/*  You should have received a copy of the GNU Lesser General Public */
/*  License along with this program.  If not, see */
/*  <http://www.gnu.org/licenses/>. */

/*  routines in this file return success codes */



package wb;

import static wb.Ents.*;
import static wb.Blkio.*;
import static wb.Blink.*;
import static wb.Wbdefs.*;
import static wb.Handle.*;
import static wb.Wbsys.*;
import static wb.Blk.*;
import static wb.BooleanMethods.*;
import static wb.Lck.*;
import static wb.Ent.*;
import static wb.Seg.*;
import static wb.Han.*;
public class Segs {

/* The @var{bsiz} of a segment (given in call to @code{make_seg}) */
/* is a parameter crucial for performance; balancing CPU time traversing */
/* blocks with file-system latency.  @var{bsiz} should be an */
/* integer multiple of the file-system block size. */
/* In the 1990s our nominal @var{bsiz} was 2.kiB; now it should */
/* probably be 4.kiB, 8.kiB, or 16.kiB. */

public static int segCntr = 4;

public static Seg segChain = null;

public static Lck segChainLck = makeLck( -3);

/*  Create segment-descriptor and splice into seg-chain. */

public static Seg newSeg(String name)
{
  if (!(a2b(lckTab)))
    {
      dprintf(">>>>ERROR<<<< "+(name)+": WB not initialized (need to call "+("initWb")+")\n");
      return null;
    }
  else {
    lck(segChainLck);
    {
      Seg seg = newSegd(segCntr);
      if (a2b(segChain))
	{
	  Seg prevSeg = seg_Prv(segChain);
	  seg_SetNxt(prevSeg, seg);
	  seg_SetPrv(seg, prevSeg);
	  seg_SetNxt(seg, segChain);
	  seg_SetPrv(segChain, seg);
	}
      else {
	seg_SetNxt(seg, seg);
	seg_SetPrv(seg, seg);
      }
      seg_SetFlc(seg, new int [flcLen]);
      seg_SetFlcLen(seg, 0);
      segCntr = 1+(segCntr);
      segChain = seg;
      unlck(segChainLck);
      return seg;
    }
  }
}

/*  lock must already be held on seg-chain-lck */

public static void freeSeg(Seg seg)
{
  if (!(a2b(segChain)))
    {
      dprintf(">>>>ERROR<<<< "+("freeSeg")+": "+("segChain")+" already empty!\n");
    }
  else if ((seg_Prv(seg))==(seg))
    {
      if ((seg)!=(segChain))
	dprintf(">>>>ERROR<<<< "+("freeSeg")+": corrupted "+("segChain")+".\n");
      segChain = null;
    }
  else {
    Seg nextSeg = seg_Nxt(seg);
    Seg prevSeg = seg_Prv(seg);
    seg_SetNxt(prevSeg, nextSeg);
    seg_SetPrv(nextSeg, prevSeg);
    if ((seg)==(segChain))
      segChain = ((segChain)==(nextSeg)
	?null
	:nextSeg);
  }
}


public static boolean checkFile_P(String filename)
{
  java.io.RandomAccessFile file = blkio_OpenReadOnlyFile(filename, 0x80);
  byte []tblk = new byte[0x19];
  subbytesRead(tblk, 0, 0x19, file);
  {
    boolean allok_P = 0==(str2long(tblk, 0))
    && 0==(str2long(tblk, 4))
    && 0==(str2long(tblk, 8))
    && 0x77==((tblk[0x17] & 0xFF))
    && 0x62==((tblk[0x18] & 0xFF));
    blkio_FileClose(file);
    return allok_P;
  }
}

/*  */
/*  */
/* Opens the database file @var{filename} and returns a @dfn{seg}, false otherwise. */
/* The database will be read-only if the @var{mutable_P} argument is false.  It will */
/* be read-write if the @var{mutable_P} argument is true. */

public static Seg openSeg(String filename,boolean mutable_P)
{
  int bsiz = 3*0x80;
  if ((bsiz)>(blkSize))
    {
      dprintf(">>>>ERROR<<<< unsupported "+("bsiz")+" "+(bsiz)+" > "+(blkSize)+"\n");
      return null;
    }
  else if (!(checkFile_P(filename)))
    return null;
  else {
    Seg seg = newSeg("openSeg");
    String reason;
Berrout:while (true) {
      if (!(a2b(seg)))
	return seg;
      else {
	java.io.RandomAccessFile file = (mutable_P
	  ?blkio_OpenModifyFile(filename, bsiz)
	  :blkio_OpenReadOnlyFile(filename, bsiz));
Lloop: while (true) {
	if ((mutable_P
	      ?outputPort_P(file)
	      :inputPort_P(file)))
	  {
	    seg_SetPort(seg, file);
	    seg_SetBsiz(seg, bsiz);
	    seg_SetUsed(seg, 2);
	    seg_SetStr(seg, filename);
	    seg_SetFlcLen(seg, (mutable_P
		? -1
		: -2));
	    {
	      Han han = seg_RtHan(seg);
	      byte []tmpStr = new byte[5];
	      if (err_P(btOpen(seg, 0, han, (wcbSap)+(wcbSar))))
		{
		  reason = "bt-open 0";
		  break Berrout;
		}
	      else if (2!=(btGet(han, bsizByts, 4, tmpStr)))
		{
		  reason = "BSIZ";
		  break Berrout;
		}
	      else if ((bsiz)!=(str2short(tmpStr, 0)))
		{
		  blkio_FileClose(file);
		  bsiz = str2short(tmpStr, 0);
		  if ((bsiz)>(blkSize))
		    {
		      reason = "BSIZ too big.";
		      break Berrout;
		    }
		  else {
		    file = (mutable_P
		      ?blkio_OpenModifyFile(filename, bsiz)
		      :blkio_OpenReadOnlyFile(filename, bsiz));
		    continue Lloop;
		  }
		}
	      else if (4!=(btGet(han, usedByts, 4, tmpStr)))
		{
		  reason = "USED";
		  break Berrout;
		}
	      else {
		seg_SetUsed(seg, str2long(tmpStr, 0));
		if (5!=(btGet(han, fldByts, 3, tmpStr)))
		  {
		    reason = "FLD";
		    break Berrout;
		  }
		else if (err_P(btOpen(seg, str2long(tmpStr, 1), seg_FlHan(seg), wcbSar)))
		  {
		    reason = "FLC";
		    break Berrout;
		  }
		else {
		  if ((han_Typ(seg_FlHan(seg)))!=(frlTyp))
		    dprintf("Older type freelist - still supported.\n");
		  han_SetWcb(seg_FlHan(seg), wcbSar);
		  return seg;
		}
	      }
	    }
	  }
	else {
	  if ((mutable_P
		?inputPort_P(file)
		:outputPort_P(file)))
	    blkio_FileClose(file);
	  dprintf(">>>>ERROR<<<< could not open file "+(filename)+"\n");
	  freeSeg(seg);
	  return null;
	}
	}
      }
    }
    dprintf(">>>>ERROR<<<< not a database "+(filename)+" "+(reason)+"\n");
    if (a2b(seg))
      {
	blkio_FileClose(seg_Port(seg));
	doSegBuffers(seg, getMethod("wb.Ents", "recycleEnt", new Class[]{Ent.class}));
	seg_SetPort(seg, null);
	seg_SetStr(seg, null);
	seg_SetUsed(seg, 0);
	freeSeg(seg);
      }
    return null;
  }
}

/*  */
/* Closes database segment @var{seg} and the file containing it.  If @var{hammer_P} is */
/* NULL, then if there are any problems freeing buffers, then the close */
/* is aborted.  A status code is returned. */

public static int closeSeg(Seg seg,boolean hammer_P)
{
  return closeSegd(seg, hammer_P, !((true)));
}


public static int closeSegd(Seg seg,boolean hammer_P,boolean all_P)
{
  if (!(all_P))
    lck(segChainLck);
  if ((!(a2b(seg_Str(seg)))
      || 0==(seg_Used(seg))))
    {
      if (!(all_P))
	unlck(segChainLck);
      return 0;
    }
  else {
    flushFlc(seg, 5);
    /*leave only enough blocks to fit in flc in superblock.*/
    if ((seg_FlcLen(seg))>=0)
      {
	byte []tmpStr = new byte[0x14];
	{
	  int i =  -1+(seg_FlcLen(seg));
	  while (!(0 > (i))) {
	    long2str(tmpStr, 4*(i), seg_Flc(seg)[i]);
	    {
	      i =  -1+(i);
	    }
	  }
	}
	btPut(seg_RtHan(seg), flcByts, 3, tmpStr, 4*(seg_FlcLen(seg)));
	seg_SetFlcLen(seg,  -1);
      }
    {
      int ans = doSegBuffers(seg, getMethod("wb.Ents", "flushBuffer", new Class[]{Ent.class}));
      if ((success_P(ans)
	  || hammer_P))
	{
	  if (!(success_P(ans)))
	    ans = notpres;
	  doSegBuffers(seg, getMethod("wb.Ents", "purgeBuffer", new Class[]{Ent.class}));
	  btClose(seg_RtHan(seg));
	  btClose(seg_FlHan(seg));
	  blkio_FileClose(seg_Port(seg));
	  seg_SetPort(seg, null);
	  seg_SetStr(seg, null);
	  
	  seg_SetFlc(seg, null);
	  seg_SetFlcLen(seg, 0);
	  seg_SetUsed(seg, 0);
	  freeSeg(seg);
	}
      if (!(all_P))
	unlck(segChainLck);
      return ans;
    }
  }
}


public static byte []dbVersionStr = stringToBytes("wb-2a3");

public static byte []dbAuthorsStr = stringToBytes("A. Jaffer, J. Finger, R. Zito-Wolf");

/*  */
/*  */
/* The integer @var{bsiz} specifies the size of B-tree blocks.  @var{bsiz} should be an */
/* integer multiple of the file-system block size.  Nominal value is */
/* 4096. */
/*  */
/* @code{makeSeg} returns an open new empty mutable database named backed by file */
/* @var{filename} if successful; otherwise false is returned. */

public static Seg makeSeg(String filename,int bsiz)
{
  Seg seg = newSeg("makeSeg");
  if (!(a2b(seg)))
    return seg;
  else if ((bsiz)>(blkSize))
    {
      dprintf(">>>>ERROR<<<< unsupported "+("bsiz")+" "+(bsiz)+" > "+(blkSize)+"\n");
      unlck(seg_Lck(seg));
      return null;
    }
  else {
    java.io.RandomAccessFile file = blkio_CreateFile(filename, bsiz);
    if (outputPort_P(file))
      {
	seg_SetPort(seg, file);
	seg_SetBsiz(seg, bsiz);
	seg_SetUsed(seg, 3);
	seg_SetStr(seg, filename);
	seg_SetFlcLen(seg,  -1);
	{
	  Han rtHan = seg_RtHan(seg);
	  Han han = seg_FlHan(seg);
	  byte []tmpStr = new byte[5];
	  if (a2b(btOpenNew(seg, 0, rtHan, (wcbSap)+(wcbSar), dirTyp))
	      && a2b(btOpenNew(seg, 1, han, wcbSap, dirTyp))
	      && a2b(btOpenNew(seg, 2, han, wcbSar, frlTyp)))
	    {
	      btPut(rtHan, noByts, 0, dbVersionStr, dbVersionStr.length);
	      long2str(tmpStr, 0, seg_Used(seg));
	      btPut(rtHan, usedByts, 4, tmpStr, 4);
	      short2str(tmpStr, 0, seg_Bsiz(seg));
	      btPut(rtHan, bsizByts, 4, tmpStr, 2);
	      tmpStr[0] = (byte) (0xff&4);
	      long2str(tmpStr, 1, 1);
	      btPut(rtHan, rootByts, 4, tmpStr, 5);
	      long2str(tmpStr, 1, 2);
	      btPut(rtHan, fldByts, 3, tmpStr, 5);
	      btPut(rtHan, flcByts, 3, noByts, 0);
	      if ((bsiz)>0x80)
		btPut(rtHan, authorsByts, 7, dbAuthorsStr, dbAuthorsStr.length);
	      return seg;
	    }
	  else {
	    dprintf(">>>>ERROR<<<< couldn't allocate ents for file "+(filename)+"\n");
	    unlck(seg_Lck(seg));
	    closeSeg(seg, true);
	    return null;
	  }
	}
      }
    else {
      dprintf(">>>>ERROR<<<< couldn't create new file "+(filename)+"\n");
      unlck(seg_Lck(seg));
      return null;
    }
  }
}

/* @noindent The write-control-bits argument (@var{wcb}) to these */
/* functions controls the latency of updates to the file after various */
/* operations.  These bits are defined as follows: */
/*  */
/* @multitable @columnfractions .1 .3 .6 */
/* @item @var{value} */
/* @tab C-name */
/* @tab Meaning */
/* @item 1 */
/* @tab wcb_sap */
/* @tab save block after PUTs */
/* @cindex wcb_sap */
/* @item 2 */
/* @tab wcb_sar */
/* @tab save block after REMOVEs */
/* @cindex wcb_sar */
/* @item 4 */
/* @tab wcb_sac */
/* @cindex wcb_sac */
/* @tab force block save after cached block changes */
/* @item 8 */
/* @tab wcb_fac */
/* @tab flush buffer entirely after cached block changes (not currently */
/* implemented) */
/* @cindex wcb_fac */
/* @end multitable */
/* Opens bt-handle @var{han} to seg number @var{seg}, block number @var{blkNum}, and */
/* returns the type of the block.  If no such block exists or is not a */
/* root block, then a (negative) status code is returned. */

public static int btOpen(Seg seg,int blkNum,Han han,int wcb)
{
  if (a2b(seg)
      && a2b(seg_Str(seg)))
    {
      Ent ent = getEnt(seg, blkNum, accread);
      if (!(a2b(ent)))
	return argerr;
      else if (!(root_P(ent_Blk(ent))))
	{
	  releaseEnt(ent, accread);
	  dprintf(">>>>ERROR<<<< "+("btOpen")+": not a root "+(seg_Id(seg))+":"+(blkNum)+"\n");
	  return argerr;
	}
      else {
	btOpenInitHan(han, ent, wcb);
	releaseEnt(ent, accread);
	return han_Typ(han);
      }
    }
  else return argerr;
}


public static void btOpenInitHan(Han han,Ent ent,int wcb)
{
  int typ = blk_Typ(ent_Blk(ent));
  Seg seg = ent_Seg(ent);
  int blkNum = ent_Id(ent);
  han_SetSeg(han, seg);
  han_SetNum(han, blkNum);
  han_SetLast(han, blkNum);
  han_SetTyp(han, typ);
  if ((typ)==(dirTyp))
    wcb = (wcb)|((wcbSap)+(wcbSar));
  han_SetWcb(han, wcb);
}

/*  Does calling GET-BUK-WAIT admit the possiblity of deadly embrace? */

public static int btOpenNew(Seg seg,int blkNum,Han han,int wcb,int typ)
{
  if (0 > (blkNum))
    {
      dprintf(">>>>ERROR<<<< negative block number "+(blkNum)+"\n");
      return 0;
    }
  else {
    Ent ent = tryGetFreeEnt(null,  -1);
    while (!(a2b(ent))) {
      {
	ent = tryGetFreeEnt(null,  -1);
      }
    }
    initLeafBlk(ent_Blk(ent), blkNum, typ);
    blk_SetTime(ent_Blk(ent), time(longZero));
    ent_SetNext(ent, getBukWait(seg_Id(seg), blkNum));
    setBuk(seg_Id(seg), blkNum, ent);
    ent_SetAcc(ent, accwrite);
    ent_SetSeg(ent, seg);
    ent_SetId(ent, blkNum);
    ent_SetAge(ent,  -127);
    ent_SetDty(ent, true);
    ent_SetPus(ent, 0);
    ent_SetRef(ent, 1);
    relBuk(seg_Id(seg), blkNum);
    btOpenInitHan(han, ent, wcb);
    ents_EntWrite(ent);
    releaseEnt(ent, accwrite);
    return 1;
  }
}

/* Creates a new root block in seg @var{seg} of type @var{typ}, opens bt-handle */
/* @var{han} to it, and returns a status code.  If @var{seg} has insufficient room to */
/* create a new tree, then the @var{noroom} status code is returned. */
/*  */
/* @code{btCreate} can be used to create temporary b-trees.  Temporary trees will be */
/* be reclaimed by check program after system crashes.  In order to */
/* make a tree persistent, add it to a directory (tree). */

public static int btCreate(Seg seg,int typ,Han han,int wcb)
{
  Ent ent = createNewBlkEnt(seg);
  if (a2b(ent))
    {
      int blkNum = ent_Id(ent);
      initLeafBlk(ent_Blk(ent), blkNum, typ);
      ent_SetPus(ent, 0);
      btOpenInitHan(han, ent, wcb);
      ents_EntWrite(ent);
      releaseEnt(ent, accwrite);
      return success;
    }
  else return noroom;
}

/* Closes bt-handle @var{han} and returns @var{SUCCESS}. */
/*  */
/* Currently, @code{btClose} has no effect other than to clear @var{han}. */

public static int btClose(Han han)
{
  han_SetSeg(han, null);
  han_SetNum(han, 0);
  han_SetTyp(han, 0);
  han_SetLast(han, 0);
  return success;
}

}
